#!/usr/bin/env python3

"""Must be run from directory containing edgelist_cyclo.csv,
edgelist_anti and SHP-triplet directory for interpolated eddies.

"""

import matplotlib.pyplot as plt
import networkx as nx
import itertools
import numpy as np
from matplotlib import patches
import cartopy.crs as ccrs

color_iter = itertools.cycle(('#1f77b4', '#aec7e8', '#ff7f0e',
                              '#ffbb78', '#2ca02c', '#98df8a',
                              '#d62728', '#ff9896', '#9467bd',
                              '#c5b0d5', '#8c564b', '#c49c94',
                              '#e377c2', '#f7b6d2', '#7f7f7f',
                              '#c7c7c7', '#bcbd22', '#dbdb8d',
                              '#17becf', '#9edae5'))

def assign_all_components(G):
    G.graph["component_list"] = list(nx.weakly_connected_components(G))

    for component in G.graph["component_list"]:
        for n in component: G.nodes[n]["component"] = component

def is_node_type(G, n, label):
    if label == "root":
        return G.in_degree[n] == 0
    elif label == "leaf":
        return G.out_degree[n] == 0
    elif label == "merge":
        return G.in_degree[n] >= 2
    elif label == "split":
        return G.out_degree[n] >= 2
    elif label == "all":
        return True
    elif isinstance(label, int):
        return n[0] == label

def plot_nbunch(G, nbunch, color = '#1f78b4', label = None, ax = None):
    if ax is None: ax = plt.gca()
    pos = G.nodes.data("coordinates")
    xy = np.asarray([pos[n] for n in nbunch])
    src_crs = ccrs.PlateCarree()
    ax.scatter(xy[:, 0], xy[:, 1], s = 10, c = color, marker='o',
               transform = src_crs)
    
    for e in G.edges(nbunch):
        arrow = patches.FancyArrowPatch(pos[e[0]], pos[e[1]],
                                        arrowstyle = '-|>', color = color,
                                        mutation_scale = 10,
                                        transform = src_crs)
        ax.add_patch(arrow)

    if label is not None:
        for n in nbunch:
            if is_node_type(G, n, label):
                xy = ax.projection.transform_point(*pos[n], src_crs)
                ax.annotate(str(n), xy, color = color, xytext = (2, 2),
                            textcoords = 'offset points')

def plot_all_components(G, label):
    for component, color in zip(G.graph["component_list"], color_iter):
        plot_nbunch(G, component, color, label)
    
def plot_descendants(G, n, label):
    nbunch = nx.descendants(G, n) | {n}
    plot_nbunch(G, nbunch, label = label)

def animate_nbunch(G, nbunch):
    sorted_nbunch = sorted(nbunch)
    j = 0
    date = sorted_nbunch[0][0]
    
    while j < len(sorted_nbunch):
        # {sorted_nbunch[j][0] == date}
        j += 1
        while j < len(sorted_nbunch) and sorted_nbunch[j][0] == date: j += 1
        plt.clf()
        plot_nbunch(G, sorted_nbunch[:j], label = date)
        plt.waitforbuttonpress()
        date += 1
    
if __name__ == "__main__":
    import report_graph
    import argparse
    import plot_eddy_contours
    import time
    
    parser = argparse.ArgumentParser(description = __doc__)
    parser.add_argument("shp_tr_dir", help = "directory containing triplet of "
                        "shapefiles, with visible eddies at all dates")
    parser.add_argument("-l", "--label",
                        choices = ["root", "leaf", "split", "merge"])
    parser.add_argument("-s", "--save", metavar = "format",
                        help = "Save file to specified format")
    parser.add_argument("-t", "--time", action = "store_true",
                        help = "Report elapsed time")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-n", "--node", help = "Select component containing "
                       "node, with orientation cyclo or anti", nargs = 3,
                       metavar = ("date_ind", "eddy_ind", "orient"))
    group.add_argument("-w", "--window", help = "choose a limited plot window",
                       type = float, nargs = 4,
                       metavar = ("lllon", "lllat", "urlon", "urlat"))
    args = parser.parse_args()

    plt.figure()

    if args.window is not None:
        if args.window[2] - args.window[0] > 360:
            sys.exit("bad values of urlon and lllon")

        if args.time:
            print("Reading edge lists in current directory...")
            t0 = time.perf_counter()
            
        G_cyclo = report_graph.read_eddy_graph("edgelist_cyclo.csv",
                                               args.shp_tr_dir)
        G_anti = report_graph.read_eddy_graph("edgelist_anti.csv",
                                               args.shp_tr_dir)

        if args.time:
            t1 = time.perf_counter()
            print("Elapsed time:", t1 - t0, "s")
            t0 = t1
            print("Finding components...")

        for G in [G_cyclo, G_anti]:
            for n, d in G.nodes.items():
                if plot_eddy_contours.in_window(d["coordinates"], args.window):
                    if "component" not in d: report_graph.add_component(G, n)

        if args.time:
            t1 = time.perf_counter()
            print("Elapsed time:", t1 - t0, "s")
            t0 = t1
            print("Plotting...")
            
        dest_crs = ccrs.PlateCarree((args.window[0] + args.window[2]) / 2)
        ax = plt.axes(projection = dest_crs)
        for G in [G_cyclo, G_anti]: plot_all_components(G, args.label)
    elif args.node is not None:
        source = (int(args.node[0]), int(args.node[1]))

        if args.time:
            print("Reading edge list in current directory...")
            t0 = time.perf_counter()
            
        G = report_graph.read_eddy_graph(f"edgelist_{args.node[2]}.csv",
                                         args.shp_tr_dir)

        if args.time:
            t1 = time.perf_counter()
            print("Elapsed time:", t1 - t0, "s")
            t0 = t1
            print("Finding component...")
            
        report_graph.add_component(G, source)

        if args.time:
            t1 = time.perf_counter()
            print("Elapsed time:", t1 - t0, "s")
            print("Plotting...")
            
        dest_crs = ccrs.PlateCarree(G.nodes[source]["coordinates"][0])
        ax = plt.axes(projection = dest_crs)
        plot_nbunch(G, G.nodes[source]["component"], label = args.label)
    else:
        if args.time:
            print("Reading edge lists in current directory...")
            t0 = time.perf_counter()
            
        G_cyclo = report_graph.read_eddy_graph("edgelist_cyclo.csv",
                                               args.shp_tr_dir)
        G_anti = report_graph.read_eddy_graph("edgelist_anti.csv",
                                               args.shp_tr_dir)

        if args.time:
            t1 = time.perf_counter()
            print("Elapsed time:", t1 - t0, "s")
            t0 = t1
            print("Finding components...")
            
        for G in [G_cyclo, G_anti]: assign_all_components(G)

        if args.time:
            t1 = time.perf_counter()
            print("Elapsed time:", t1 - t0, "s")
            print("Plotting...")
            
        ax = plt.axes(projection = ccrs.PlateCarree())
        for G in [G_cyclo, G_anti]: plot_all_components(G, args.label)

    ax.coastlines()
    ax.gridlines(draw_labels = True)
    if args.time: print("Elapsed time:", time.perf_counter() - t1, "s")

    if args.save:
        plt.savefig(f"plot_traj.{args.save}")
        print(f'Created "plot_traj.{args.save}".')
    else:
        plt.show()
