#!/usr/bin/env python3

"""Converts adt_YYYY-MM-DD.mat files to a triplet of shapefiles. The
    mat files are assumed to be in v7.3 (HDF5) format and will be
    converted to v6 format before being read. The data in each input
    file is the set of detected instantaneous eddies at a given date.

"""

import shapefile
import numpy as np
from os import path
import scipy.io as sio
import datetime
import f90nml
import sys
import pty
import os

def define_fields(writers):
    """writers is a dictionary of shapefile.Writer objects."""
    
    writers["extr"].field("ssh", "N", 13, 6)
    writers["extr"].field("date_index", "N", 5)
    writers["extr"].field("eddy_index", "N", 5)
    writers["extr"].field("speed", "N", 13, 6)

    writers["center"].field("date_index", "N", 5)
    writers["center"].field("eddy_index", "N", 5)

    writers["outer"].field("r_eq_area", "N", 10, 4)
    writers["outer"].field("ssh", "N", 13, 6)
    writers["outer"].field("date_index", "N", 5)
    writers["outer"].field("eddy_index", "N", 5)

    writers["max_speed"].field("r_eq_area", "N", 10, 4)
    writers["max_speed"].field("ssh", "N", 13, 6)
    writers["max_speed"].field("date_index", "N", 5)
    writers["max_speed"].field("eddy_index", "N", 5)

def write(writers, cell, cyclone):
    """writers is a dictionary. cell is a numpy.ndarray with shape (number
    of eddies, 21). cyclone is a boolean. The eddy identification
    numbers start at 1 and increment by 1 for each eddy in cell.

    """
    eddy_index = 0

    for eddy in cell:
        # eddy is a numpy.ndarray with shape (21,). The
        # description of each element of eddy is given by
        # matlab_data["Fields"]. Note that amplitudes in
        # eddy[7] and eddy[11] are positive regardless of eddy
        # orientation.
        eddy_index += 1

        writers["extr"].point(eddy[0], eddy[1])
        i = np.argwhere(matlab_data["X"] == eddy[0]).item()
        j = np.argwhere(matlab_data["Y"] == eddy[1]).item()
        speed = 1e4 if np.isnan(eddy[12]) else eddy[12]
        writers["extr"].record(ssh = matlab_data["ADT"][i, j],
                               date_index = date_index, eddy_index = eddy_index,
                               speed = speed)

        if cyclone:
            ssh = matlab_data["ADT"][i, j] + eddy[7]
        else:
            ssh =  matlab_data["ADT"][i, j] - eddy[7]

        writers["center"].point(eddy[2], eddy[3])
        writers["center"].record(date_index = date_index,
                                 eddy_index = eddy_index)

        writers["outer"].record(r_eq_area = eddy[6], ssh = ssh, 
                                date_index = date_index,
                                eddy_index = eddy_index)
        polyline = np.stack((eddy[4], eddy[5]), axis = 1)
        writers["outer"].poly([polyline])

        if cyclone:
            ssh = matlab_data["ADT"][i, j] + eddy[11]
        else:
            ssh =  matlab_data["ADT"][i, j] - eddy[11]

        writers["max_speed"].record(r_eq_area = eddy[10], ssh = ssh, 
                                    date_index = date_index,
                                    eddy_index = eddy_index)
        polyline = np.stack((eddy[8], eddy[9]), axis = 1)
        writers["max_speed"].poly([polyline])

    n_shapes = len(writers["extr"])
    writers["ishape_last"].write(str(n_shapes - 1) + "\n")

if len(sys.argv) == 1: sys.exit("Required argument: first input .mat file")
adt_file = sys.argv[1]
dirname, basename = path.split(adt_file)
my_date = datetime.datetime.strptime(basename, "adt_%Y-%m-%d.mat").date()

if len(sys.argv) == 3:
    final_date = datetime.datetime.strptime(sys.argv[2], "%Y-%m-%d").date()
else:
    final_date = my_date

factories = [dict(writers = {}, cell_name = "Anticyclonic_Cell",
                  cyclone = False, SHP_coll = "Snapshot_anti"),
             dict(writers = {}, cell_name = "Cyclonic_Cell", cyclone = True,
                  SHP_coll = "Snapshot_cyclo")]

for f in factories:
    file = path.join(f["SHP_coll"], "extremum")
    f["writers"]["extr"] = shapefile.Writer(file, shapeType = shapefile.POINT)
    
    file = path.join(f["SHP_coll"], "center")
    f["writers"]["center"] = shapefile.Writer(file, shapeType = shapefile.POINT)
    
    file = path.join(f["SHP_coll"], "outermost_contour")
    f["writers"]["outer"] = shapefile.Writer(file,
                                             shapeType = shapefile.POLYGON)
    
    file = path.join(f["SHP_coll"], "max_speed_contour")
    f["writers"]["max_speed"] = shapefile.Writer(file,
                                                 shapeType = shapefile.POLYGON)
    
    define_fields(f["writers"])
    file = path.join(f["SHP_coll"], "ishape_last.txt")
    f["writers"]["ishape_last"] = open(file, "w")
    

while True:
    os.symlink(adt_file, "adt.mat")
    pty.spawn(["matlab", "-nojvm", "-r", "inst_eddies"])
    matlab_data = sio.loadmat("adt_v6.mat", squeeze_me = True,
                              variable_names = ("date_num", "X", "Y",
                                                "Anticyclonic_Cell",
                                                "Cyclonic_Cell", "ADT"))
    # matlab_data is a dictionary.

    assert datetime.date.fromordinal(matlab_data["date_num"] - 366) == my_date
    date_index = my_date - datetime.date(1950, 1, 1)
    date_index = date_index.days

    for f in factories:
        write(f["writers"], matlab_data[f["cell_name"]], f["cyclone"])
    
    os.remove("adt.mat")
    my_date += datetime.timedelta(1)
    if my_date > final_date: break
    basename = my_date.strftime("adt_%Y-%m-%d.mat")
    adt_file = path.join(dirname, basename)
    
nml = f90nml.Namelist()
nml["grid_nml"] = dict(corner_deg = [matlab_data["X"][0], matlab_data["Y"][0]],
                       nlon = matlab_data["X"].size,
                       nlat = matlab_data["Y"].size)

for f in factories:
    nml.write(path.join(f["SHP_coll"], "grid_nml.txt"), force = True)
    for k, v in f["writers"].items(): v.close()

os.remove("adt_v6.mat")
